/*
 * wheelfix  -- fix wheel by post WM_MOUSEWHEEL event to
 *              under window of pointer
 *
 * $Id: wheelfix.c,v 1.2 2003/07/28 07:48:59 hos Exp $
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <strings.h>
#include <errno.h>
#include <windows.h>
#include "wheelhook.h"
#include "resource.h"


#define WM_TASKTRAY (WM_APP+0x10)


const char *window_class_name = "wheelfix";
const char *window_title_name = "wheelfix";

const int notify_icon_id = 1;

static int taskbar_created_message = 0;

static HINSTANCE instance = NULL;
static HMENU popup_menu = NULL;

static int pause = 0;



struct uint_ptr_pair {
    unsigned int key;
    void *ptr;
};

typedef LRESULT (*msg_func_t)(HWND, UINT, WPARAM, LPARAM);


static void *assq_pair(const struct uint_ptr_pair *pair,
                       unsigned int key, void *def_ptr)
{
    while(pair->ptr != NULL) {
        if(pair->key == key) {
            return pair->ptr;
        }

        pair++;
    }

    return def_ptr;
}



/* task tray operation */
static int set_tasktray_icon(HWND wnd, int msg)
{
    NOTIFYICONDATA notify_icon;
    HICON icon;

    icon = LoadIcon(instance, "TRAY_ICON");
    if(icon == NULL) {
        return 0;
    }

    memset(&notify_icon, 0, sizeof(notify_icon));
    notify_icon.cbSize = sizeof(notify_icon);
    notify_icon.hWnd = wnd;
    notify_icon.uID = notify_icon_id;
    notify_icon.uFlags = NIF_ICON | NIF_MESSAGE | NIF_TIP;
    notify_icon.uCallbackMessage = WM_TASKTRAY;
    notify_icon.hIcon = icon;
    strcpy(notify_icon.szTip, window_title_name);

    return Shell_NotifyIcon(msg, &notify_icon);
}


/* pause menu item operation */
static int set_pause_menu_item(int state)
{
    MENUITEMINFO item_info;

    memset(&item_info, 0, sizeof(item_info));
    item_info.cbSize = sizeof(item_info);
    item_info.fMask = MIIM_STATE;
    item_info.fState = state;

    return SetMenuItemInfo(popup_menu, ID_MENU_PAUSE, FALSE, &item_info);
}



/* error message */
static void error_message(const char *msg)
{
    char buf[1024];

    sprintf(buf, "%s: %lu", msg, GetLastError());
    MessageBox(NULL, buf, window_title_name, MB_OK | MB_ICONERROR);
}


/* warning message with last error number */
static void warning_message_le(const char *msg)
{
    fprintf(stderr, "warning: %s (%lu)\n", msg, GetLastError());
    SetLastError(ERROR_SUCCESS);
}



/* menu item: pause */
static LRESULT pause_proc(HWND wnd, UINT msg, WPARAM wparam, LPARAM lparam)
{
    int ret;

    /* toggle pause state */
    pause = !pause;

    set_pause_menu_item(pause ? MFS_CHECKED : MFS_UNCHECKED);

    if(pause != 0) {
        ret = clear_wheel_fix_hook();
        if(ret == 0) {
            error_message("clear_wheel_fix_hook() failed");
        }
    } else {
        ret = set_wheel_fix_hook();
        if(ret == 0) {
            error_message("set_wheel_fix_hook() failed");
        }
    }

    return 0;
}


/* menu item: exit */
static LRESULT exit_proc(HWND wnd, UINT msg, WPARAM wparam, LPARAM lparam)
{
    DestroyWindow(wnd);

    return 0;
}


/* default message: wrap difference declare (stdcall and cdecl) */
static LRESULT default_proc(HWND wnd, UINT msg, WPARAM wparam, LPARAM lparam)
{
    return DefWindowProc(wnd, msg, wparam, lparam);
}




/* WM_CREATE */
static LRESULT create_proc(HWND wnd, UINT msg, WPARAM wparam, LPARAM lparam)
{
    int ret;

    /* load menu */
    {
        HMENU menu;
        menu = LoadMenu(instance, "TRAY_MENU");
        if(menu == NULL) {
            error_message("LoadMenu() failed");

            PostQuitMessage(0);
            return 0;
        }

        popup_menu = GetSubMenu(menu, 0);
        if(popup_menu == 0) {
            error_message("GetSubMenu() failed");

            PostQuitMessage(0);
            return 0;
        }
    }

    /* add icon to task tray */
    ret = set_tasktray_icon(wnd, NIM_ADD);
    if(ret == 0) {
        error_message("set_tasktray_icon() failed");

        PostQuitMessage(0);
        return 0;
    }

    /* set wheelfix hook */
    ret = set_wheel_fix_hook();
    if(ret == 0) {
        error_message("set_wheel_fix_hook() failed");

        PostQuitMessage(0);
        return 0;
    }

    return 0;
}


/* WM_DESTROY */
static LRESULT destroy_proc(HWND wnd, UINT msg, WPARAM wparam, LPARAM lparam)
{
    int ret;

    ret = set_tasktray_icon(wnd, NIM_DELETE);
    if(ret == 0) {
        error_message("set_tasktray_icon() failed");
    }

    ret = DestroyMenu(popup_menu);
    if(ret == 0) {
        error_message("DestroyMenu() failed");
    }

    ret = clear_wheel_fix_hook();
    if(ret == 0) {
        error_message("clear_wheel_fix_hook() failed");
    }

    PostQuitMessage(0);

    return 0;
}


/* WM_COMMAND */
static LRESULT command_proc(HWND wnd, UINT msg, WPARAM wparam, LPARAM lparam)
{
    struct uint_ptr_pair msg_func[] = {
        {ID_MENU_PAUSE, pause_proc},
        {ID_MENU_EXIT, exit_proc},
        {0, NULL}
    };
    msg_func_t func;

    func = (msg_func_t)assq_pair(msg_func, wparam, default_proc);

    return func(wnd, msg, wparam, lparam);
}


/* WM_TASKTRAY */
static LRESULT tasktray_proc(HWND wnd, UINT msg, WPARAM wparam, LPARAM lparam)
{
    if(lparam != WM_RBUTTONDOWN) {
        return 0;
    }

    {
        POINT pos;
        int ret;

        GetCursorPos(&pos);

        SetForegroundWindow(wnd);
        SetFocus(wnd);

        ret = TrackPopupMenu(popup_menu, TPM_LEFTALIGN | TPM_BOTTOMALIGN,
                             pos.x, pos.y, 0, wnd, NULL);
        if(ret == 0) {
            error_message("TrackPopupMenu() failed");
        }
    }

    return 0;
}


/* "TaskbarCreated" */
static LRESULT taskbar_created_proc(HWND wnd, UINT msg,
                                    WPARAM wparam, LPARAM lparam)
{
    int ret;

    ret = set_tasktray_icon(wnd, NIM_ADD);
    if(ret == 0) {
        warning_message_le("set_tasktray_icon() failed");
    }

    return 0;
}


/* process message */
static LRESULT CALLBACK window_procedure(HWND wnd, UINT msg,
                                         WPARAM wparam, LPARAM lparam)
{
    struct uint_ptr_pair msg_func[] = {
        {0, taskbar_created_proc},
        {WM_CREATE, create_proc},
        {WM_DESTROY, destroy_proc},
        {WM_COMMAND, command_proc},
        {WM_TASKTRAY, tasktray_proc},
        {0, NULL}
    };
    msg_func_t func;

    msg_func[0].key = taskbar_created_message;
    func = (msg_func_t)assq_pair(msg_func, msg, default_proc);

    return func(wnd, msg, wparam, lparam);
}


/* create window */
static HWND create_window(void)
{
    int ret;

    {
        WNDCLASSEX wc;

        memset(&wc, 0, sizeof(wc));
        wc.cbSize = sizeof(wc);
        wc.style = 0;
        wc.lpfnWndProc = window_procedure;
        wc.cbClsExtra = 0;
        wc.cbWndExtra = 0;
        wc.hInstance = instance;
        wc.hIcon = NULL;
        wc.hCursor = NULL;
        wc.hbrBackground = (HBRUSH)(COLOR_WINDOW + 1);
        wc.lpszMenuName = NULL;
        wc.lpszClassName = window_class_name;
        wc.hIconSm = NULL;

        ret = RegisterClassEx(&wc);
        if(ret == 0) {
            return NULL;
        }
    }

    {
        HWND wnd;

        wnd = CreateWindow(window_class_name, window_title_name,
                           WS_OVERLAPPED,
                           CW_USEDEFAULT, CW_USEDEFAULT,
                           CW_USEDEFAULT, CW_USEDEFAULT,
                           NULL, NULL, instance, NULL);

        return wnd;
    }
}



/* main message loop */
static int message_loop(void)
{
    MSG msg;
    int loop, ret;

    loop = 1;
    while(loop) {
        ret = GetMessage(&msg, NULL, 0, 0);
        if(ret < 0) {
            error_message("GetMessage() failed");
            return 1;
        }

        if(ret != 0) {
            TranslateMessage(&msg);
            DispatchMessage(&msg);
        }

        loop = ret;
    }

    return msg.wParam;
}


/* entry point of program */
int main(int ac, char **av)
{
    HWND wnd;
    int ret;

    {
        HMODULE module;

        module = GetModuleHandle(NULL);
        if(module == NULL) {
            error_message("GetModuleHandle() failed");
            exit(1);
        }

        instance = (HINSTANCE)module;
    }

    taskbar_created_message = RegisterWindowMessage("TaskbarCreated");
    if(taskbar_created_message == 0) {
        error_message("RegisterWindowMessage() failed");
        exit(1);
    }

    wnd = create_window();
    if(wnd == NULL) {
        error_message("create_window() failed");
        exit(1);
    }

    ShowWindow(wnd, SW_HIDE);
    
    ret = UpdateWindow(wnd);
    if(ret == 0) {
        error_message("UpdateWindow() failed");
        exit(1);
    }

    ret = message_loop();

    return ret;
}
